(* +--------------------------------------------------------------+
   |                                                              |
   |                    Analyseur d'expression arithmtiques      |
   |                                                              |
   +--------------------------------------------------------------+ *)

open Printf
open Numerix

module M(E:Int_type) : sig

  (* value une chaine de caractres *)
  exception Incorrect of string
  val evalue : string -> E.t

end = struct

  exception Incorrect of string
  open E

  type lexme =
    | Nombre of t            (* nombre                *)
    | Op1  of op1            (* op. unaire prfixe    *)
    | Op2  of op2            (* op. binaire infixe    *)
    | Op12 of op1*op2        (* op. unaire ou binaire *)
    | Parg                   (* parenthse ouvrante   *)
    | Pard                   (* parenthse fermante   *)
  and op1 = t -> t
  and op2 = int * (t -> t -> t) (* priorit, fonction *)


  let dix = of_int(10)
  let puiss a b = pow a (int_of b)

  (* dcompose une chane de caractres en liste de lexmes *)
  let dcompose s =
    let l = ref [] in
    for i=0 to String.length(s)-1 do match s.[i] with
      | ' ' -> ()
      | '~' -> l := Op1(neg)          :: !l
      | '+' -> l := Op2(1,add)        :: !l
      | '-' -> l := Op12(neg,(1,sub)) :: !l
      | '*' -> l := Op2(2,mul)        :: !l
      | '/' -> l := Op2(2,quo)        :: !l
      | '^' -> l := Op2(3,puiss)      :: !l
      | '(' -> l := Parg              :: !l
      | ')' -> l := Pard              :: !l
      | '0'..'9' as c -> begin
	  let x = of_int(Char.code(c) - Char.code('0')) in
	  match !l with
	    | Nombre(y)::suite when (i > 0) & (s.[i-1] <> ' ')
		-> l := Nombre(add x (mul dix y))::suite
	    | _ -> l := Nombre(x) :: !l
	end
      | _ -> raise (Incorrect s)
    done;
    List.rev !l


  (* rduction d'une expression infixe *)
  let rec empile pile x = match x,pile with

    (* opration mixte *)
    | Op12(op1,op2), Nombre(_)::_ -> empile pile (Op2(op2))
    | Op12(op1,op2), _            -> empile pile (Op1(op1))

    (* opration unaire en attente *)
    | Nombre(n), Op1(g)::suite ->
	empile suite (Nombre(g n))

    (* opration binaire prioritaire *)
    | Op2(p,g),  Nombre(b)::Op2(q,h)::Nombre(a)::suite when p <= q ->
	empile (Nombre(h a b)::suite) x

    (* parenthse fermante *)
    | Pard,  Nombre(b)::Op2(q,h)::Nombre(a)::suite ->
	empile (Nombre(h a b)::suite) x
    | Pard, Nombre(n)::Parg::suite ->
	empile suite (Nombre n)

    (* autres cas *)
    | _,_ -> x::pile

  (* valuation d'une chane *)
  let evalue s =
    let pile = List.fold_left empile [Parg] (dcompose s) in
    match empile pile Pard with
      | [Nombre(x)] -> x
      | _ -> raise (Incorrect s)

end
